/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include "rumd_stats.h"

#include "rumd/EnergiesMetaData.h"
#include "rumd/RUMD_Error.h"
#include <cstring>
#include <cassert>
#include <cerrno>
#include <cmath>
#include <zlib.h>
#include <iostream>
#include <sstream>
#include <iomanip>
#include <cstdio>

rumd_stats::rumd_stats() : directory("TrajectoryFiles"), 
			   column_labels(),
			   verbose(false),
			   mean_vals(0),
			   meanSq_vals(0),
			   num_cols(0),
			   count(0),
			   EMD(),
			   numFilenameDigits(4) {}


void rumd_stats::Allocate(unsigned int set_num_cols) {
  if (num_cols != set_num_cols) {
    Free();
    num_cols = set_num_cols;
    mean_vals = new double[num_cols];
    meanSq_vals = new double[num_cols];
    
  }
}

void rumd_stats::Free()
  {
    if(mean_vals) delete [] mean_vals;
    if(meanSq_vals) delete [] meanSq_vals;
  }


void rumd_stats::ComputeStats (unsigned int first_block, int last_block) {
  unsigned int last_saved_block;
  unsigned int last_block_to_read;
  
  gzFile gzfp;
  FILE *fp;
  
  std::string lastComp_filename = directory+"/LastComplete_energies.txt";
  fp=fopen(lastComp_filename.c_str(),"r");
  if (!fp)
    throw RUMD_Error("rumd_stats","ComputeStats",std::string("Error opening file ")+lastComp_filename + ": " + strerror(errno));

  fscanf(fp, "%u", &last_saved_block);
  fclose(fp);
  if(verbose) std::cout << "last completed energy block is " << last_saved_block << std::endl;

  if(first_block > last_saved_block) {
    std::cerr << "Warning: user-specified first_block out of range; setting to zero" << std::endl;
    first_block = 0;
  }
  if(last_block < 0)
    last_block_to_read = last_saved_block;
  else
    if((unsigned int)last_block > last_saved_block) {
      last_block_to_read = last_saved_block;
      std::cerr << "Warning, specified last_block out of range; using last saved block" << std::endl;
    }
    else
      last_block_to_read = (unsigned int) last_block;

  if(first_block > last_block_to_read)
    std::cerr << "Warning: last_block smaller than first_block: no statistics computed" << std::endl;

  //read energy file for first block 
  
  std::ostringstream E0_Name;
  E0_Name << directory << "/energies"<< std::setfill('0') << std::setw(numFilenameDigits) << std::setprecision(numFilenameDigits) << first_block << ".dat.gz";

  
  if(verbose) std::cout << "Reading block " << first_block << ": " << E0_Name.str() << std::endl << std::endl;
  gzfp=gzopen(E0_Name.str().c_str(),"r");
  if (!gzfp)
    throw RUMD_Error("rumd_stats","ComputeStats",std::string("Error opening file ")+E0_Name.str() + ": " + strerror(errno));
  
  char strWorkspace[256];
  unsigned int maxLen = sizeof(strWorkspace)/sizeof(char);
  // read the first line
  bool ok = gzgets(gzfp, strWorkspace, maxLen);
  if(!ok)
    throw RUMD_Error("rumd_stats","ComputeStats",std::string("Error reading first line of file")+E0_Name.str());
  else {
    // replace final newline with a zero
    strWorkspace[strlen(strWorkspace)-1] = 0;
  }

  EMD.ReadMetaData(strWorkspace, verbose);
  
  gzclose(gzfp);

  int set_num_cols = 0;
  count = 0;
  column_labels.clear();

  std::map<std::string, bool>::const_iterator it;
  for(it = EMD.Start(); it != EMD.End(); it++ )
    if(EMD.Get(it->first))  {
      set_num_cols += 1;
      column_labels.push_back(EMD.GetFileString(it->first));
    }
  if(verbose) std::cout << "Number of columns is " << set_num_cols << std::endl;
  
  Allocate(set_num_cols);
  
  memset(mean_vals, 0, num_cols*sizeof(double));
  memset(meanSq_vals, 0, num_cols*sizeof(double));

  // now read all blocks
  for (unsigned int i=first_block; i<=last_block_to_read; i++) {
    std::ostringstream E_Name;
    E_Name << directory << "/energies"<< std::setfill('0') << std::setw(numFilenameDigits) << std::setprecision(numFilenameDigits) << i << ".dat.gz";

    if(verbose) std::cout << "Reading block " << i << ": " << E_Name.str() << std::endl;
    gzfp=gzopen(E_Name.str().c_str(),"r");
    if (!gzfp)
      throw RUMD_Error("rumd_stats","ComputeStats",std::string("Error opening file ") + E_Name.str() + ": " + strerror(errno));

    // read first line, but don't do anything with it (could perhaps
    // check for consistency ... )
    ok = gzgets(gzfp, strWorkspace, maxLen);
    if(!ok)
      throw RUMD_Error("rumd_stats","ComputeStats",std::string("Error reading first line of file ")+E_Name.str());

    // read data lines
    ok = gzgets(gzfp, strWorkspace, maxLen);
    while(ok) {
      int offset = 0;
      int partial_offset = 0;
      double value = 0.;
      for(unsigned int idx = 0; idx < num_cols; ++idx) {
	sscanf(strWorkspace+offset, "%lf%n", &value, &partial_offset);
        offset += partial_offset;
	mean_vals[idx] += value;
	meanSq_vals[idx] += value*value;
      } 
      count += 1;
      ok = gzgets(gzfp, strWorkspace, maxLen);
    } // end while loop over lines in this file

    gzclose(gzfp);
  } // end loop over blocks

  if(count)
    for(unsigned int idx = 0; idx < num_cols; ++idx) {
      mean_vals[idx] /= count;
      meanSq_vals[idx] /= count;
    }
}


void rumd_stats::PrintStats() {
  if(!mean_vals || !meanSq_vals)
    throw RUMD_Error("rumd_stats","PrintStats","Must call ComputeStats first");
  std::cout << "Total line count: " << count << std::endl;
  std::cout << "quantity, mean value, variance, standard deviation" << std::endl;

  std::map<std::string, bool>::const_iterator it;
  it = EMD.Start(); 
  while(!it->second) { it++;}
  for(unsigned int idx = 0; idx < num_cols; ++idx) {
    double variance = (meanSq_vals[idx] - mean_vals[idx]*mean_vals[idx]);//*count/(count-1.);
    double std_dev = sqrt(variance);
    printf("%s\t%12.6g\t%12.6g\t%12.6g\n", EMD.GetFileString(it->first).c_str(), mean_vals[idx], variance, std_dev );
    do { it++; } while (it != EMD.End() && !it->second);
  }
  assert(it == EMD.End());
}



void rumd_stats::WriteStats() {
  if(!mean_vals || !meanSq_vals)
    throw RUMD_Error("rumd_stats","PrintStats","Must call ComputeStats first");

  std::map<std::string, bool>::const_iterator it;

  FILE* meanFile = fopen("energies_mean.dat","w");
  FILE* meanSqFile = fopen("energies_mean_sq.dat","w");
  FILE* varFile = fopen("energies_var.dat","w");
  
  fprintf(meanFile,"# ");
  fprintf(meanSqFile,"# ");
  fprintf(varFile,"# ");	
  for(it = EMD.Start(); it != EMD.End(); it++) 
    if(it->second){
      fprintf(meanFile," %s",EMD.GetFileString(it->first).c_str());
      fprintf(meanSqFile," %s",EMD.GetFileString(it->first).c_str());
      fprintf(varFile," %s",EMD.GetFileString(it->first).c_str());   
    }
  fprintf(meanFile,"\n");
  fprintf(meanSqFile,"\n");
  fprintf(varFile,"\n");

  it = EMD.Start(); 
  while(!it->second) it++;
  for(unsigned int idx = 0; idx < num_cols; ++idx) {
    double variance = (meanSq_vals[idx] - mean_vals[idx]*mean_vals[idx]);//*count/(count-1.);
    fprintf(meanFile," %12.10g",mean_vals[idx]);
    fprintf(meanSqFile," %12.10g",meanSq_vals[idx]);
    fprintf(varFile," %12.10g",variance);
    do { it++; } while (it != EMD.End() && !it->second);
  }
  fprintf(meanFile,"\n");
  fprintf(varFile,"\n");
  fclose(meanFile);
  fclose(meanSqFile);
  fclose(varFile);
  assert(it == EMD.End());
}
