#ifndef PAIRPOTENTIAL_H
#define PAIRPOTENTIAL_H

/*
  Copyright (C) 2010  Thomas Schrøder
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  LICENSE.txt file for license details.
*/

////////////////////////////////////////////////////////////////////////////////////////////////////////
// This header implements specialized Pair Potentials. Implement your potential here.
// DISCLAIMER: (Rcut_ij)^2 in absolute units should ALWAYS be the 0'th parameter ie. h_params[i][j][0].
////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <cstdio>
#include <cmath>
#include <cstring>
#include <vector>
#include <algorithm>

#include "rumd/RUMD_Error.h"
#include "rumd/Potential.h"
#include "rumd/NeighborList.h"
#include "rumd/SimulationBox.h"

class PairPotential : public Potential{
 private:
  PairPotential(const PairPotential&);
  PairPotential& operator=(const PairPotential&); 
  virtual void LaunchPE_Kernel()=0;
  static std::vector<unsigned int> objectIndexList;
  static unsigned int GetNewObjectIndex() {
    unsigned int currentSize = objectIndexList.size();
    unsigned int smallest_not_there = currentSize;
    if(currentSize > 0 && currentSize != objectIndexList[currentSize-1] + 1) {
      unsigned int index = 0;
      while(index == objectIndexList[index])
	index++;
      smallest_not_there = index;
    }

    objectIndexList.push_back(smallest_not_there);
    sort(objectIndexList.begin(), objectIndexList.end());
    return smallest_not_there;
  }
  static void RemoveObjectIndex(unsigned int index) {
    
    std::vector<unsigned int>::iterator found_index = std::find(objectIndexList.begin(), objectIndexList.end(), index);
    if(found_index != objectIndexList.end())
      objectIndexList.erase(found_index);
    else
      throw RUMD_Error("PairPotential","RemoveIndex","Could not find index for pair potential object");
  }
protected:
  const unsigned int pp_objectIndex;
  float potential_factor;
  float h_params[MaxNumTypes][MaxNumTypes][NumParam];
  NeighborList neighborList;
  RectangularSimulationBox* testRSB;
  float4* d_f_pe;
  float4* h_f_pe;
  unsigned int allocatedSize;

  void AllocatePE_Array(unsigned int nvp);

public: 
  PairPotential() : pp_objectIndex(GetNewObjectIndex()), potential_factor(1.0), neighborList(), testRSB(0), d_f_pe(0), h_f_pe(0), allocatedSize(0) { 
    memset(h_params, 0, MaxNumTypes * MaxNumTypes * NumParam * sizeof(float));
    if(pp_objectIndex >= MaxNumPP_Objects)
      throw RUMD_Error("PairPotential","PairPotential","Maximum number of pair potential objects has been reached.");
    SetID_String("pair_pot");
  }
  virtual ~PairPotential() { RemoveObjectIndex(pp_objectIndex); }

  // Needed due to no virtual function on device.
  virtual float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w) { return 0.f; };  
  void WritePotentials(SimulationBox* simBox);

  void CopyParamsToGPU(); 
  void SaveNeighborList(){ neighborList.SaveNeighborList(); };
  void RestoreNeighborList(){ neighborList.RestoreNeighborList(); };
  void ResetNeighborList(){ neighborList.ResetNeighborList(); };
  void PrintExclusionList(){ neighborList.PrintExclusionList(); };
 
  // Set methods.
  void SetNbListSkin( float skin ){ neighborList.SetCutoffs(this, skin); }
  void SetExclusionList( NBTYPEv* list ){ neighborList.SetExclusionList(list); };
  void SetSampleParameters( KernelPlan& kPlan, ParticleData* pd, SimulationBox* simBox, float4* d_misc );
  void SetExclusionBond( int1 *h_btlist, uint1* h_blist, unsigned num_part, unsigned num_virt_part, 
			 unsigned max_num_bonds, unsigned etype );
  void SetExclusionMolecule(int1 *h_mlist, unsigned molindex, unsigned max_num_uau);

  // Get methods.
  float GetPotentialParameter(int type1, int type2, int paramIdx) const{ return h_params[type1][type2][paramIdx]; }
  float GetNbListSkin(){ return neighborList.GetSkin(); }
  double GetPotentialEnergy();
};

////////////////////////////////////////////////////////////////////////////////////////////////
// Standard 12-6 Lennard-Jones (cut and shifted potential).
// v(r) =  4 Epsilon ( (Sigma/r)^12 -    (Sigma/r)^6 ) - v_cut.
// s(r) = 48 Epsilon ( (Sigma/r)^14 - 0.5(Sigma/r)^8 )/Sigma^2 = -r^-1*dv/dr (F_ij = s * r_ij).
////////////////////////////////////////////////////////////////////////////////////////////////

class Pot_LJ_12_6 : public PairPotential{
 public: 
  Pot_LJ_12_6(){ potential_factor = 1.0 / 12.0; SetID_String("potLJ");}
  void CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs);
  void LaunchPE_Kernel();
  float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w){ return ComputeInteraction(dist2, param, my_f, my_w); };       

  Potential* Copy() {
    Pot_LJ_12_6* new_pot = new Pot_LJ_12_6();
    for(unsigned int i = 0; i < MaxNumTypes; i++)
      for(unsigned int j = 0; j < MaxNumTypes; j++) {
	if(h_params[i][j][1] > 0.) {
	  float Sigma = 1./sqrt(h_params[i][j][1]);
	  float Rcut = sqrt(h_params[i][j][0])/Sigma;
	  float Epsilon = h_params[i][j][2]/48.f;
	  new_pot->SetParams(i, j, Sigma, Epsilon, Rcut);
	}
      }
    return new_pot;
  }

  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut) {
    if( i >= MaxNumTypes || j >= MaxNumTypes || NumParam <= 4 )
      throw( RUMD_Error("PairPotential","Pot_LJ_12_6","The potential could not be allocated" ) );
    
    h_params[i][j][0] = Rcut * Rcut * Sigma * Sigma;                             // (Rcut_ij)^2
    h_params[i][j][1] = 1.f / ( Sigma * Sigma );                                 // (1/Sigma_ij)^2
    h_params[i][j][2] = 48.f * Epsilon;                                          // Prefactor for forces
    h_params[i][j][3] = 48.f * Epsilon * ( pow(Rcut,-12.f) - pow(Rcut,-6.f) );   // Potential at Rcut
    if(verbose)
      printf("Pot_LJ_12_6: %f %f %f %f\n", Sigma, Epsilon, Rcut, h_params[i][j][3]);
    CopyParamsToGPU();
  }                                    
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w){
    float invDist2 = 1.0f / ( dist2 * param[1] );                                // (sigma/r)^2 
    float invDist6 = invDist2 * invDist2 * invDist2;                             // (sigma/r)^6 
    float s = param[1] * param[2] * invDist6 * invDist2 * ( invDist6 - 0.5f );   // F_ij = s * r_ij
    (*my_f).w += param[2] * invDist6 * ( invDist6 - 1.0f ) - param[3];           // U_ij/potential_factor 
    (*my_w).w += param[2] * invDist6 * ( invDist6 - 0.5f );                      // W_ij (virial)
    return s;
  }
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w, float4* my_misc){ return 0.f; }
};

////////////////////////////////////////////////////////////////////////////////////////////////
// Standard 12-6 Lennard-Jones (cut and shifted force).
// v(r) =  4 Epsilon ( (Sigma/r)^12 - (Sigma/r)^6 ) + ( r - r_cut ) f_cut - v_cut.
// s(r) = 48 Epsilon ( (Sigma/r)^14 - 0.5(Sigma/r)^8 )/Sigma^2 - f_cut / r. 
////////////////////////////////////////////////////////////////////////////////////////////////

class Pot_LJ_12_6_SF : public PairPotential{
 public: 
  Pot_LJ_12_6_SF(){ potential_factor = 1.0 / 12.0; SetID_String("potLJSF"); }
  void CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs);
  void LaunchPE_Kernel();
  float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w){ return ComputeInteraction(dist2, param, my_f, my_w); };       
  
  Potential* Copy(){
    Pot_LJ_12_6_SF* new_pot = new Pot_LJ_12_6_SF();
    for(unsigned int i = 0; i < MaxNumTypes; i++)
      for(unsigned int j = 0; j < MaxNumTypes; j++) {
	if(h_params[i][j][1] > 0.) {
	  float Sigma = 1./sqrt(h_params[i][j][1]);
	  float Rcut = sqrt(h_params[i][j][0])/Sigma;
	  float Epsilon = h_params[i][j][2]/48.f;
	  new_pot->SetParams(i, j, Sigma, Epsilon, Rcut);
	}
      }
    return new_pot;
  }
  
  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut) {
    if( i >= MaxNumTypes || j >= MaxNumTypes || NumParam <= 7 )
      throw( RUMD_Error("PairPotential","Pot_LJ_12_6_SF","The potential could not be allocated" ) );
    
    h_params[i][j][0] = Rcut * Rcut * Sigma * Sigma;                                                    // (Rcut_ij)^2
    h_params[i][j][1] = 1.f / ( Sigma * Sigma );                                                        // (1/Sigma_ij)^2
    h_params[i][j][2] = 48.f * Epsilon;                                                                 // Prefactor for forces
    h_params[i][j][3] = 48.f * Epsilon * ( pow(Rcut,-12.f) - pow(Rcut,-6.f) );                          // Potential at Rcut
    h_params[i][j][4] = 48.f * Epsilon * ( pow(Rcut,-13.f) / Sigma - 0.5f * pow(Rcut,-7.f) / Sigma );   // Force at Rcut
    h_params[i][j][5] = 12.f * h_params[i][j][4];                                                       // Force at Rcut with potential prefactor
    h_params[i][j][6] = Rcut * Sigma;                                                                   // Rcut
    if(verbose)
      printf("Pot_LJ_12_6_SF: %f %f %f %f %f\n", Sigma, Epsilon, Rcut, h_params[i][j][3], h_params[i][j][4]);
    CopyParamsToGPU();
  }                                    
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w){
    float dist = sqrtf(dist2);
    float invDist2 = 1.0f / ( dist2 * param[1] );                                                               // (sigma/r)^2 
    float invDist6 = invDist2 * invDist2 * invDist2;                                                            // (sigma/r)^6 
    float s = param[1] * param[2] * invDist6 * invDist2 * ( invDist6 - 0.5f ) - param[4] / dist;                // F_ij = s * r_ij
    (*my_f).w += param[2] * invDist6 * ( invDist6 - 1.0f ) + ( dist - param[6] ) * param[5] - param[3];         // U_ij / potential_factor 
    (*my_w).w += param[2] * invDist6 * ( invDist6 - 0.5f ) - dist * param[4];                                   // W_ij (virial)
    return s;
  }
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w, float4* my_misc){ return 0.f; }
};

////////////////////////////////////////////////////////////////////////////////////////////////
// Standard 12-6 Lennard-Jones (Gromacs smoothing of force).
// v(r) = ?
// s(r) = ?
////////////////////////////////////////////////////////////////////////////////////////////////

class Pot_LJ_12_6_Gromacs : public PairPotential{
 public: 
  Pot_LJ_12_6_Gromacs(){ potential_factor = 1.0 / 12.0; SetID_String("potLJGr");}

  void CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs);
  void LaunchPE_Kernel();
  
  float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w){ return ComputeInteraction(dist2, param, my_f, my_w); };   

  Potential* Copy() {
    Pot_LJ_12_6_Gromacs* new_pot = new Pot_LJ_12_6_Gromacs();
    for(unsigned int i = 0; i < MaxNumTypes; i++)
      for(unsigned int j = 0; j < MaxNumTypes; j++) {
	if(h_params[i][j][1] > 0.) {
	  float Sigma = 1./sqrt(h_params[i][j][1]);
	  float Rcut = sqrt(h_params[i][j][0])/Sigma;
	  float Epsilon = h_params[i][j][2]/48.f;
	  new_pot->SetParams(i, j, Sigma, Epsilon, Rcut);
	}
      }
    return new_pot;
  }

  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut){
    if( i >= MaxNumTypes || j >= MaxNumTypes || NumParam <= 9 )
      throw( RUMD_Error("PairPotential","Pot_12_6_Gromacs","The potential could not be allocated" ) );
    
    // Hard-coded.
    float Rsmooth = 2.45;
    float diff = Sigma * ( Rcut - Rsmooth );
    
    float Force = ( 48.f * Epsilon * ( pow(Rcut,-13.f) - 0.5 * pow(Rcut,-7.f) ) ) / Sigma;   
    float Deriv = ( - 624.f * Epsilon * pow(Rcut,-14.f) + 168 * Epsilon * pow(Rcut,-8.f) ) / ( Sigma * Sigma ); 
    
    // Gromacs constants.
    float A = ( - 3.f * Force + Deriv * diff ) / powf(diff,2.f);
    float B = ( - Deriv * diff + 2.f * Force ) / powf(diff,3.f);
    float C = ( A / 3.f ) * powf(diff,3.f) + ( B / 4.f ) * powf(diff,4.f);
    
    // GPU parameters.
    h_params[i][j][0] = Rcut * Rcut * Sigma * Sigma;                           
    h_params[i][j][1] = 1.f / ( Sigma * Sigma );                               
    h_params[i][j][2] = 48.f * Epsilon;                                        
    h_params[i][j][3] = 48.f * Epsilon * ( pow(Rcut,-12.f) - pow(Rcut,-6.f) ); 
    h_params[i][j][4] = Rsmooth * Rsmooth * Sigma * Sigma;
    h_params[i][j][5] = A;
    h_params[i][j][6] = B;
    h_params[i][j][7] = 12.f * C;
    h_params[i][j][8] = Rsmooth * Sigma;
    if(verbose)
      printf("Pot_LJ_12_6_Gromacs: %f %f %f %f - A = %f; B = %f; C = %f;\n", Sigma, Epsilon, Rcut, Rsmooth, A, B, C);
    CopyParamsToGPU();
  }
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w){ return 0.f; }
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w, float4* my_misc){
    
    float invDist2 = 1.0f / ( dist2 * param[1] );                               
    float invDist6 = invDist2 * invDist2 * invDist2;                           
    float common =  param[1] * param[2] * invDist6 * invDist2;
    float s = common * ( invDist6 - 0.5f ); 
    
    (*my_f).w += param[2] * invDist6 * ( invDist6 - 1.0f ) - param[3] + param[7]; // U_ij
    (*my_w).w += param[2] * invDist6 * ( invDist6 - 0.5f );                       // W_ij 
    (*my_misc).x += 3.f * s + common * ( -14.f * invDist6 + 4.f );                // T_config

    // r >= Rsmooth.
    if( dist2 >= param[4] ){
      float r = sqrtf(dist2); 
      float diff = r - param[8]; 
      float diff_sq = diff * diff;
 
      float pforce = ( diff_sq * ( param[5] + param[6] * diff ) );
      float force = pforce / r; 
      s += force;
      
      (*my_f).w += diff_sq * ( - 4.f * diff * param[5] - 3.f * diff_sq * param[6] );    // U_ij 
      (*my_w).w += r * pforce;                                                          // W_ij
      (*my_misc).x += 2.f * force + diff * ( 2.f * param[5] + 3.f * diff * param[6] );  // T_config
    }
    return s;
  }
};

////////////////////////////////////////////////////////////////////////////////////////////////
// Generalized Lennard-Jones (cut and shifted potential).
// v(r) = Epsilon/(m -n) ( n*(Sigma/r)^m -  m*(Sigma/r)^n ) - v_cut.
// s(r) = ?.
////////////////////////////////////////////////////////////////////////////////////////////////

class Pot_gLJ_m_n : public PairPotential{
 public: 
  float m, n; // exponents 
 
  Pot_gLJ_m_n(float set_m, float set_n): m(set_m), n(set_n){ potential_factor = 1.0f / (m*n); SetID_String("pot_gLJ");}

  float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w) { return ComputeInteraction(dist2, param, my_f, my_w); };       
  void CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs);
  void LaunchPE_Kernel();

  Potential* Copy() {
    Pot_gLJ_m_n* new_pot = new Pot_gLJ_m_n(m, n);
    for(unsigned int i = 0; i < MaxNumTypes; i++)
      for(unsigned int j = 0; j < MaxNumTypes; j++) {
	if(h_params[i][j][1] > 0.) {
	  float Sigma = 1./sqrt(h_params[i][j][1]);
	  float Rcut = sqrt(h_params[i][j][0])/Sigma;
	  float Epsilon = h_params[i][j][2]/(m*n / (m-n));
	  new_pot->SetParams(i, j, Sigma, Epsilon, Rcut);
	}
      }
    return new_pot;
  }

  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut){
    if( i >= MaxNumTypes || j >= MaxNumTypes || NumParam <= 6 )
      throw( RUMD_Error("PairPotential","Pot_gLJ_m_n","The potential could not be allocated" ) );
    
    h_params[i][j][0] = Rcut * Rcut * Sigma * Sigma;                             // (Rcut_ij)^2
    h_params[i][j][1] = 1.f / ( Sigma * Sigma );                                 // (1/Sigma_ij)^2
    h_params[i][j][2] = m*n / (m-n) * Epsilon;                                   // Prefactor for forces
    h_params[i][j][3] = n*powf(Rcut,-m) - m*powf(Rcut,-n);                       // "Potential" at Rcut
    h_params[i][j][4] = m*0.5f;                                                  // m/2.0
    h_params[i][j][5] = n*0.5f;                                                  // n/2.0
    if(verbose)
      printf("Pot_gLJ_m_n: %f %f %f %f %f %f\n", m, n, Sigma, Epsilon, Rcut, h_params[i][j][3]);
    CopyParamsToGPU();
  }
 
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w) {
    float invDist2 = 1.0f / ( dist2 * param[1] );                             
    float invDist_m = exp(log(invDist2)*param[4]); 
    float invDist_n = exp(log(invDist2)*param[5]);
    float s =   param[1] * param[2] * (invDist_m - invDist_n) * invDist2; 
    (*my_f).w += param[2] * (2.0f*param[5]*invDist_m - 2.0f*param[4]*invDist_n - param[3]);       // U_ij/potential_factor   
    (*my_w).w += param[2] * (  invDist_m -   invDist_n); 
    return s;
  }
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w, float4* my_misc){ return 0.f; }
};

////////////////////////////////////////////////////////////////////////////////////////////////
// Standard 12-6 Lennard-Jones (cut and shifted potential with re-def. Sigma HO = 2^(1/6) Sigma LJ).
// v(r) =    Epsilon ( (Sigma/r)^12 -  2 *  (Sigma/r)^6 ) - v_cut.
// s(r) = 12 Epsilon ( (Sigma/r)^14 - (Sigma/r)^8 )/Sigma^2 = -r^-1*dv/dr (F_ij = s * r_ij).
////////////////////////////////////////////////////////////////////////////////////////////////

class Pot_LJ_12_6_NEW : public PairPotential{
 public: 
  Pot_LJ_12_6_NEW(){ potential_factor = 1.0 / 12.0; SetID_String("potLJN");}

  void CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs);
  void LaunchPE_Kernel();

  float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w){ return ComputeInteraction(dist2, param, my_f, my_w); };

  Potential* Copy() {
    Pot_LJ_12_6_NEW * new_pot = new Pot_LJ_12_6_NEW ();
    for(unsigned int i = 0; i < MaxNumTypes; i++)
      for(unsigned int j = 0; j < MaxNumTypes; j++) {
	if(h_params[i][j][1] > 0.) {
	  float Sigma = 1./sqrt(h_params[i][j][1]);
	  float Rcut = sqrt(h_params[i][j][0])/Sigma;
	  float Epsilon = h_params[i][j][2]/12.f;
	  new_pot->SetParams(i, j, Sigma, Epsilon, Rcut);
	}
      }
    return new_pot;
  }
  
  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut){
    if( i >= MaxNumTypes || j >= MaxNumTypes || NumParam <= 4 )
      throw( RUMD_Error("PairPotential","Pot_LJ_12_6_NEW","The potential could not be allocated" ) );
    
    h_params[i][j][0] = Rcut * Rcut * Sigma * Sigma ;                             // (Rcut_ij)^2
    h_params[i][j][1] =  1.f / ( Sigma * Sigma );                                // (1/Sigma_ij)^2
    h_params[i][j][2] = 12.f * Epsilon;                                          // Prefactor for forces
    h_params[i][j][3] = 12.f * Epsilon * ( pow(Rcut,-12.f) - 2.f * pow(Rcut,-6.f) );   // Potential at Rcut
    if(verbose)
      printf("Pot_LJ_12_6_NEW: %f %f %f %f\n", Sigma, Epsilon, Rcut, h_params[i][j][3]);
    CopyParamsToGPU();
  }                                    

  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w){
    float invDist2 = 1.0f / ( dist2 * param[1] );                                // (sigma/r)^2 
    float invDist6 = invDist2 * invDist2 * invDist2;                             // (sigma/r)^6 
    float s =   param[1] * param[2] * invDist6 * invDist2 * ( invDist6 - 1.f ); // F_ij = s * r_ij
    (*my_f).w += param[2] * invDist6 * ( invDist6 - 2.0f ) - param[3];           // U_ij/potential_factor 
    (*my_w).w += param[2] * invDist6 * ( invDist6 - 1.0f );                      // W_ij (virial)
    return s;
  }
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w, float4* my_misc){ return 0.f; }
};

////////////////////////////////////////////////////////////////////////////////////////////////
// 12-6 Lennard-Jones plus Gaussian potential (cut and shifted potential).
// v(r) = Epsilon ( (Sigma/r)^12 -  2(Sigma/r)^6 - Epsilon_0 (exp[-(r - r_0)^2 / (2 Sigma_0^2)] ) ) - v_cut.
// s(r) = ?.
////////////////////////////////////////////////////////////////////////////////////////////////

class Pot_LJ_12_6_Gauss : public PairPotential{
 public: 
  Pot_LJ_12_6_Gauss(){ potential_factor = 1.0 / 12.0; SetID_String("potLJG");}

  void CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs);
  void LaunchPE_Kernel();

  float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w){ return ComputeInteraction(dist2, param, my_f, my_w); };
  
  Potential* Copy() {
    Pot_LJ_12_6_Gauss* new_pot = new Pot_LJ_12_6_Gauss();
    for(unsigned int i = 0; i < MaxNumTypes; i++)
      for(unsigned int j = 0; j < MaxNumTypes; j++) {
	if(h_params[i][j][1] > 0.) {
	  float Sigma = 1./sqrt(h_params[i][j][1]);
	  float Rcut = sqrt(h_params[i][j][0])/Sigma;
	  float Epsilon = h_params[i][j][2]/12.f;
	  float Sigma0 =  sqrt(0.5f*h_params[i][j][4]);
	  float Epsilon0 = h_params[i][j][6];
	  float r0 = h_params[i][j][5];
	  new_pot->SetParams(i, j, Sigma, Epsilon, Rcut, Sigma0, Epsilon0, r0);
	}
      }
    return new_pot;  
  }
  
  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut, float Sigma0, float Epsilon0, float r0) {
    if( i >= MaxNumTypes || j >= MaxNumTypes || NumParam <= 7 )
      throw( RUMD_Error("PairPotential","Pot_LJ_12_6_Gauss","The potential could not be allocated" ) );
    
    h_params[i][j][0] = Rcut * Rcut * Sigma * Sigma;                              // (Rcut_ij)^2  (Rcut := rcut/Sigma)
    h_params[i][j][1] = 1.f / ( Sigma * Sigma );                                 // (1/Sigma_ij)^2
    h_params[i][j][2] = 12.f * Epsilon;                                           // Prefactor for forces 
    h_params[i][j][3] = 12.f * Epsilon * ( pow (Rcut,-12.f) - 2.f *  pow (Rcut,-6.f) - (Epsilon0 * exp (- pow ((Rcut * Sigma - r0),2.f) / ( 2.f * pow (Sigma0,2.f))))); // Potential at Rcut
    h_params[i][j][4] = 2.0f * Sigma0 * Sigma0;                                   // Sigma0
    h_params[i][j][5] = r0;                                                       // r0
    h_params[i][j][6] = Epsilon0;                                                 // Epsilon0
    if(verbose)
      printf("LJ_12_6_Gauss: %f %f %f %f %f %f %f\n", Sigma, Epsilon, Rcut, h_params[i][j][3], Sigma0, r0, Epsilon0);
    CopyParamsToGPU();
  }                                    
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w) {
    float invDist2 = 1.0f / ( dist2 * param[1] );                                // (sigma/r)^2 
    float invDist6 = invDist2 * invDist2 * invDist2;                             // (sigma/r)^6 
    float r = sqrtf(dist2);
    float exp_factor = exp (- ((r - param[5]) * (r - param[5])) / param[4]);
    float s = param[1] * param[2] * invDist6 * invDist2 * ( invDist6 - 1.f ) - (param[6] * param[2] * ( r - param[5] ) / ( 6.0f * r *  param[4])) * exp_factor; 
    (*my_f).w += param[2] * (invDist6 * ( invDist6 - 2.0f ) - param[6] * exp_factor ) - param[3];                                                   // U_ij
    (*my_w).w += param[2] * invDist6 * ( invDist6 - 1.0f ) - (param[6] * param[2] * ( dist2 - r * param[5] ) / ( 6.0f *  param[4])) * exp_factor ;  // W_ij (virial)
    return s;
  }
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w, float4* my_misc){ return 0.f; }
};

////////////////////////////////////////////////////////////////////////////////////////////////
// Gaussian core potential.
// v(r) = Epsilon * exp[-(r/Sigma)^2].
// s(r) = ( 2 * Epsilon / Sigma^2 ) * exp[-(r/Sigma)^2] = -r^-1*dv/dr (F_ij = s * r_ij).
////////////////////////////////////////////////////////////////////////////////////////////////

class Pot_Gauss : public PairPotential {
 public: 

  Pot_Gauss(){ potential_factor = 1.0 / 1.0; SetID_String("potGau");}

  void CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs);
  void LaunchPE_Kernel();

  float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w){ return ComputeInteraction(dist2, param, my_f, my_w); };

  Potential* Copy() {
    Pot_Gauss* new_pot = new Pot_Gauss();
    for(unsigned int i = 0; i < MaxNumTypes; i++)
      for(unsigned int j = 0; j < MaxNumTypes; j++) {
	if(h_params[i][j][1] > 0.) {
	  float Sigma = 1./sqrt(h_params[i][j][1]);
	  float Rcut = sqrt(h_params[i][j][0])/Sigma;
	  float Epsilon = h_params[i][j][2];
	  new_pot->SetParams(i, j, Sigma, Epsilon, Rcut);
	}
      }
    return new_pot;
  }

  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut){
    if( i >= MaxNumTypes || j >= MaxNumTypes || NumParam <= 4 )
      throw( RUMD_Error("PairPotential","Pot_Gauss","The potential could not be allocated" ) );

    h_params[i][j][0] = Rcut * Rcut * Sigma * Sigma ;                     // (Rcut_ij)^2  Rcut := rcut/Sigma
    h_params[i][j][1] = 1.f / ( Sigma * Sigma );                          // (1/Sigma_ij)^2
    h_params[i][j][2] = Epsilon;                                          // Prefactor for forces
    h_params[i][j][3] = Epsilon * exp (-( Rcut * Rcut ));                 // Potential at Rcut
    if(verbose)
      printf("Pot_Gauss: %f %f %f %f\n", Sigma, Epsilon, Rcut, h_params[i][j][3]);
    CopyParamsToGPU();
  }                                    
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w){
    float Dist2 = dist2 * param[1];                                // (r/Sigma)^2 
    float s = 2.f * param[1] * param[2] * exp(-Dist2);             // F_ij = s * r_ij
    (*my_f).w += param[2] * exp(- Dist2) - param[3];               // U_ij 
    (*my_w).w += 2.f * Dist2 * param[1] * param[2] * exp(-Dist2);  // W_ij (virial) = s * r^2
    return s;
  }

  __device__ __host__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w, float4* my_misc){ return 0.f; }
};

////////////////////////////////////////////////////////////////////////////////////////////////
// Buckingham potential.
// v(r) = Epsilon{(6/(Alpha - 6))*exp[Alpha(1-r/rm)] - (Alpha/(Alpha -6))*(rm/r)**6}.
// s(r) = -r^-1*dv/dr = ((6*Alpha*Epsilon)/(Alpha - 6)){r^(-1)*exp[Alpha(1-r/rm)] - rm^(-6)*r^(-8)}.
////////////////////////////////////////////////////////////////////////////////////////////////

class Pot_Buckingham : public PairPotential{
 public: 
  Pot_Buckingham(){ potential_factor = 1.0; SetID_String("potBuc");}

  void CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs);
  void LaunchPE_Kernel();

  float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w){ return ComputeInteraction(dist2, param, my_f, my_w); };


  Potential* Copy() {
    Pot_Buckingham* new_pot = new Pot_Buckingham();
    for(unsigned int i = 0; i < MaxNumTypes; i++)
      for(unsigned int j = 0; j < MaxNumTypes; j++) {
	if(h_params[i][j][1] > 0.) {
	  
	  float Rcut = sqrt(h_params[i][j][0]);
	  float Epsilon = h_params[i][j][1];
	  float Alpha = h_params[i][j][3];
	  float rm = h_params[i][j][4];
	  new_pot->SetParams(i, j, rm, Alpha, Epsilon, Rcut);
	}
      }
    return new_pot;
  }
  
  void SetParams(unsigned int i, unsigned int j, float rm, float Alpha, float Epsilon, float Rcut){
    if( i >= MaxNumTypes || j >= MaxNumTypes || NumParam <= 6 )
      throw( RUMD_Error("PairPotential","Pot_Buckingham","The potential could not be allocated" ) );

    h_params[i][j][0] = Rcut * Rcut;                                             // (Rcut_ij)^2
    h_params[i][j][1] = Epsilon;                                                 // Prefactor
    h_params[i][j][2] = Alpha / (Alpha - 6.f);                                   // Prefactor 
    h_params[i][j][3] = Alpha;                                                   // Prefactor
    h_params[i][j][4] = rm;                                                      // Prefactor 
    h_params[i][j][5] = Epsilon * ( (6.f/(Alpha - 6.f)) * exp (Alpha * (1.f - Rcut/rm) ) - (Alpha/(Alpha -6.f))* pow ((rm/Rcut),6.f) ); // Potential at Rcut
    if(verbose)
      printf("Pot_Buckingham: %f %f %f %f %f \n", rm, Alpha, Epsilon, Rcut, h_params[i][j][4]);
    CopyParamsToGPU();
  }                                    
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w){
    float invDist2 = 1.0f / ( dist2 );                                                      // (1/r)^2 
    float invDist6 = invDist2 * invDist2 * invDist2;                                        // (1/r)^6

    float inv_r = sqrtf (invDist2);
    float exp_factor = expf ( param[3] * (1.f - 1.f /(param[4]*inv_r)));
    float rm2 = param[4] * param[4];
    float rm6 = rm2 * rm2 * rm2;

    float s =  6.f * param[1] * param[2] * ( (inv_r/ param[4]) * exp_factor
					     - rm6 * invDist2 * invDist6);    // F_ij = s * r_ij
    (*my_f).w +=  param[1] * ((6.f*(param[2]/param[3])) * exp_factor
			      - param[2] * rm6 * invDist6) - param[5]; // U_ij
    (*my_w).w +=  6.f * param[1] * param[2]*((1.f/(param[4]*inv_r)) * exp_factor
					     -  rm6 * invDist6);             // W_ij (virial) = s * r^2
    return s;
  }
  
 __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w, float4* my_misc){ return 0.f; }
};

////////////////////////////////////////////////////////////////////////////////////////////////
// IPL potential, n=12. (cut and shifted potential).
// v(r) =    Epsilon ( (Sigma/r)^12 ) - v_cut.
// s(r) = 12 Epsilon ( (Sigma/r)^14 )/Sigma^2 = -r^-1*dv/dr (F_ij = s * r_ij).
////////////////////////////////////////////////////////////////////////////////////////////////

class Pot_IPL_12 : public PairPotential{
 public: 
  Pot_IPL_12(){ potential_factor = 1.0 / 12.0; SetID_String("potIPL12");}

  void CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs)
;
  void LaunchPE_Kernel();

  float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w){ return ComputeInteraction(dist2, param, my_f, my_w); };

  Potential* Copy() {
    Pot_IPL_12* new_pot = new Pot_IPL_12();
    for(unsigned int i = 0; i < MaxNumTypes; i++)
      for(unsigned int j = 0; j < MaxNumTypes; j++) {
	if(h_params[i][j][1] > 0.) {
	  float Sigma = 1./sqrt(h_params[i][j][1]);
	  float Rcut = sqrt(h_params[i][j][0])/Sigma;
	  float Epsilon = h_params[i][j][2]/12.f;
	  new_pot->SetParams(i, j, Sigma, Epsilon, Rcut);
	}
      }
    return new_pot;
  }

  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut) {
    if( i >= MaxNumTypes || j >= MaxNumTypes || NumParam <= 4 )
      throw( RUMD_Error("PairPotential","Pot_IPL_12","The potential could not be allocated" ) );
    
    h_params[i][j][0] = Rcut * Rcut * Sigma * Sigma;                             // (Rcut_ij)^2
    h_params[i][j][1] =  1.f / ( Sigma * Sigma );                                // (1/Sigma_ij)^2
    h_params[i][j][2] = 12.f * Epsilon;                                          // Prefactor for forces
    h_params[i][j][3] = 12.f * Epsilon * ( pow(Rcut,-12.f) );                    // Potential at Rcut
    if(verbose)
      printf("Pot_IPL_12: %f %f %f %f\n", Sigma, Epsilon, Rcut, h_params[i][j][3]);
    CopyParamsToGPU();
  }                                    
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w) {
    float invDist2  = 1.0f / ( dist2 * param[1] );                               // (sigma/r)^2 
    float invDist6  = invDist2 * invDist2 * invDist2;                            // (sigma/r)^6
    float invDist12 = invDist6 * invDist6;                                       // (sigma/r)^12    
    float s =   param[1] * param[2] * invDist12 * invDist2;                      // F_ij = s * r_ij
    (*my_f).w += param[2] * invDist12 - param[3];                                // U_ij/potential_factor 
    (*my_w).w += param[2] * invDist12;                                           // W_ij (virial)
    return s;
  }
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w, float4* my_misc){ return 0.f; }
};

////////////////////////////////////////////////////////////////////////////////////////////////
// IPL potential, n=18. (cut and shifted potential).
// v(r) =    Epsilon ( (Sigma/r)^18 ) - v_cut.
// s(r) = 18 Epsilon ( (Sigma/r)^20 )/Sigma^2 = -r^-1*dv/dr (F_ij = s * r_ij).
////////////////////////////////////////////////////////////////////////////////////////////////

class Pot_IPL_18 : public PairPotential{
 public: 
  Pot_IPL_18(){ potential_factor = 1.0 / 18.0; SetID_String("potIPL18");}

  void CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs);
  void LaunchPE_Kernel();

  float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w){ return ComputeInteraction(dist2, param, my_f, my_w); };

  Potential* Copy() {
    Pot_IPL_18* new_pot = new Pot_IPL_18();
    for(unsigned int i = 0; i < MaxNumTypes; i++)
      for(unsigned int j = 0; j < MaxNumTypes; j++) {
	if(h_params[i][j][1] > 0.) {
	  float Sigma = 1./sqrt(h_params[i][j][1]);
	  float Rcut = sqrt(h_params[i][j][0])/Sigma;
	  float Epsilon = h_params[i][j][2]/18.f;
	  new_pot->SetParams(i, j, Sigma, Epsilon, Rcut);
	}
      }
    return new_pot;
  }

  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut){
    if( i >= MaxNumTypes || j >= MaxNumTypes || NumParam <= 4 )
      throw( RUMD_Error("PairPotential","Pot_IPL_18","The potential could not be allocated" ) );

    h_params[i][j][0] = Rcut * Rcut * Sigma * Sigma;                             // (Rcut_ij)^2
    h_params[i][j][1] =  1.f / ( Sigma * Sigma );                                // (1/Sigma_ij)^2
    h_params[i][j][2] = 18.f * Epsilon;                                          // Prefactor for forces
    h_params[i][j][3] = 18.f * Epsilon * ( pow(Rcut,-18.f) );                    // Potential at Rcut
    if(verbose)
      printf("Pot_IPL_18: %f %f %f %f\n", Sigma, Epsilon, Rcut, h_params[i][j][3]);
    CopyParamsToGPU();
  }                                    

  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w) {
    float invDist2  = 1.0f / ( dist2 * param[1] );                               // (sigma/r)^2 
    float invDist6  = invDist2 * invDist2 * invDist2;                            // (sigma/r)^6
    float invDist18 = invDist6 * invDist6 * invDist6;                            // (sigma/r)^18    
    float s =   param[1] * param[2] * invDist18 * invDist2;                      // F_ij = s * r_ij
    (*my_f).w += param[2] * invDist18 - param[3];                                // U_ij/potential_factor 
    (*my_w).w += param[2] * invDist18;                                           // W_ij (virial)
    return s;
  }
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w, float4* my_misc){ return 0.f; }
};

////////////////////////////////////////////////////////////////////////////////////////////////
// IPL potential, abritrary exponent n. (cut and shifted potential).
// v(r) =   Epsilon ( (Sigma/r)^n ) - v_cut.
// s(r) = n Epsilon ( (Sigma/r)^(n+2) )/Sigma^2 = -r^-1*dv/dr (F_ij = s * r_ij).
////////////////////////////////////////////////////////////////////////////////////////////////

class Pot_IPL_n : public PairPotential{
 public: 
  float n;   // exponent

  Pot_IPL_n(): n(0.0){ potential_factor = 1.0; SetID_String("potIPLn");}

  void CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs);
  void LaunchPE_Kernel();

  float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w){ return ComputeInteraction(dist2, param, my_f, my_w); };

  Potential* Copy() {
  Pot_IPL_n* new_pot = new Pot_IPL_n();
    for(unsigned int i = 0; i < MaxNumTypes; i++)
      for(unsigned int j = 0; j < MaxNumTypes; j++) {
	if(h_params[i][j][1] > 0.) {
	  float Sigma = 1./sqrt(h_params[i][j][1]);
	  float Rcut = sqrt(h_params[i][j][0])/Sigma;
	  float Epsilon = h_params[i][j][2];
	  float n = h_params[i][j][4];
	  new_pot->SetParams(i, j, Sigma, Epsilon, Rcut, n);
	}
      }
    return new_pot;
  }

  void SetParams(unsigned int i, unsigned int j, float Sigma, float Epsilon, float Rcut, float n){
    if( i >= MaxNumTypes || j >= MaxNumTypes || NumParam <= 5 )
      throw( RUMD_Error("PairPotential","Pot_IPL_n","The potential could not be allocated" ) );

    h_params[i][j][0] = Rcut * Rcut * Sigma * Sigma;   // (Rcut_ij)^2
    h_params[i][j][1] =  1.f / ( Sigma * Sigma );      // (1/Sigma_ij)^2
    h_params[i][j][2] = Epsilon;                   // Prefactor for forces
    h_params[i][j][3] = Epsilon * ( pow(Rcut,-n) );// Potential at Rcut
    h_params[i][j][4] = n;                             // n
    if(verbose)
      printf("Pot_IPL_n=%f: %f %f %f %f\n", n, Sigma, Epsilon, Rcut, h_params[i][j][3]);
    CopyParamsToGPU();
  }                                    

  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w) {
    float invDist2   = 1.0f / ( dist2 * param[1] ); // (sigma/r)^2 
    float invDist_n = expf(logf(invDist2)*param[4]*0.5f);// (sigma/r)^n (faster)
    float s = param[1] * param[2] * param[4] * invDist_n * invDist2;// F_ij = s * r_ij
    (*my_f).w += param[2] * invDist_n - param[3];    // U_ij/potential_factor 

    (*my_w).w += param[2] * param[4] * invDist_n;// W_ij (virial)
    return s;
  }

  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w, float4* my_misc){ return 0.f; }
};

////////////////////////////////////////////////////////////////////////////////////////////////
// Dzugutov potential.
// v(r) = ?
// s(r) = ?
////////////////////////////////////////////////////////////////////////////////////////////////

class Pot_Dzugutov: public PairPotential{

 public: 
  Pot_Dzugutov(){ potential_factor = 1.f; SetID_String("potDzu");}

  void CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs);
  void LaunchPE_Kernel();

  float ComputeInteraction_cpu(float dist2, float* param, float4* my_f, float4* my_w){ return ComputeInteraction(dist2, param, my_f, my_w); };     

  Potential* Copy() {
    Pot_Dzugutov* new_pot = new Pot_Dzugutov();
    for(unsigned int i = 0; i < MaxNumTypes; i++)
      for(unsigned int j = 0; j < MaxNumTypes; j++) {
	if(h_params[i][j][1] > 0.) {
	  float a = h_params[i][j][1];
	  float b = h_params[i][j][2];
	  float c = h_params[i][j][3];
	  float d = h_params[i][j][4];
	  float A = h_params[i][j][5];
	  float B = h_params[i][j][6];
	  new_pot->SetParams(i, j, a, b, c, d, A, B);
	}
      }
    return new_pot;
  }
  
  void SetParams(unsigned int i, unsigned int j, float a, float b, float c, float d, float A, float B){
    if( i >= MaxNumTypes || j >= MaxNumTypes || NumParam <= 7 )
      throw( RUMD_Error("PairPotential","Pot_Dzugutov","The potential could not be allocated" ) );

    h_params[i][j][0] = (b - 0.000001f) * (b - 0.000001f);  
    h_params[i][j][1] = a; 
    h_params[i][j][2] = b; 
    h_params[i][j][3] = c; 
    h_params[i][j][4] = d; 
    h_params[i][j][5] = A; 
    h_params[i][j][6] = B; 
    if(verbose)
      printf("Pot_Dzugutov: a = %f, b = %f, c = %f, d = %f, A = %f, B = %f. \n", a, b, c, d, A, B);
    CopyParamsToGPU();
  }
    
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w){
    float r = sqrtf(dist2);
    float s = 0;
    
    if( r < param[2] ){ // To ensure no division by 0.
      float inv_r = 1.f / r;
      float inv_b = 1.f / ( r - param[2] );
      float exp_b = expf( param[4] * inv_b );
      float inv_b_sq_exp = inv_b * inv_b * exp_b;
      
      s = inv_r * param[4] * param[6] * inv_b_sq_exp; 
      (*my_f).w += param[6] * exp_b; // U_ij
      (*my_w).w += dist2 * s;        // W_ij
      
      if( r < param[1] ){
	float invDist2 = 1.f / dist2; 
	float invDist4 = invDist2 * invDist2;
	float invDist8 = invDist4 * invDist4;
	
	float inv_a = 1.f / ( r - param[1] );
	float exp_a = expf( param[3] * inv_a );
	
	float potential_energy = param[5] * exp_a * ( invDist8 * invDist8 - param[6] );
	float force_factor = exp_a * invDist8 * invDist8;
	float common = 16 * param[5] * force_factor * invDist2 + inv_r * param[3] * inv_a * inv_a * potential_energy;
	
	s += common;
	(*my_f).w += potential_energy; // U_ij
	(*my_w).w += dist2 * common;   // W_ij
      }
    }
    return s;
  }
  
  __host__ __device__ float ComputeInteraction(float dist2, float* param, float4* my_f, float4* my_w, float4* my_misc){ return 0; }
};

#endif // PAIRPOTENTIAL_H
