#ifndef SIMULATIONBOX_H
#define SIMULATIONBOX_H

/*
  Copyright (C) 2010  Thomas Schrøder
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  LICENSE.txt file for license details.
*/

#include "rumd/RUMD_Error.h"
#include "rumd/rumd_base.h"
#include <cmath>

// The abstract Simulation Box class. Inherit this class to build your own SimulationBox.
class SimulationBox{
 
 private:
  SimulationBox(const SimulationBox&); 
  SimulationBox& operator=(const SimulationBox&); 
  
 protected:
  float* h_boxDetails;
  float* d_boxDetails;

 public:
  SimulationBox() : h_boxDetails(0), d_boxDetails(0){};
  virtual ~SimulationBox(){};
  
  __host__ __device__ float4 calculateDistance( float4 position1, float4 position2, float* simBoxPointer );
  __host__ __device__ float4 calculateDistanceWithImages( float4 position1, float4 position2, float image1, float image2, float* simBoxPointer );
  __host__ __device__ float4 applyBoundaryCondition( float4& position, float* simBoxPointer );
  __host__ __device__ void loadMemory( float* toMem, float* fromMem );

  virtual void ScaleBox( float factor ){};

  // Set methods.
  virtual void SetBoxDetails( float* details ) = 0;

  // Get methods.
  virtual float GetVolume(){ return -1.f; };
  virtual float4 GetSimulationBox(){ float4 value = {0,0,0,0}; return value; };
  virtual SimulationBox* MakeCopy() = 0;
  virtual void CopyAnotherBox(SimulationBox* boxToBeCopied) = 0;

  virtual float* GetHostPointer(){
    if(!h_boxDetails)
      throw( RUMD_Error("SimulationBox","SimulationBox","Host pointer is invalid") );
    else
      return h_boxDetails;
  }

  virtual float* GetDevicePointer(){
    if(!d_boxDetails)
      throw( RUMD_Error("SimulationBox","SimulationBox","Device pointer is invalid") );
    else
      return d_boxDetails;
  }
};

///////////////////////////////////////////////////
// Avaliable simulation boxes
////////////////////////////////////////////////////

class RectangularSimulationBox : public SimulationBox{

 private: 
  RectangularSimulationBox(const RectangularSimulationBox&);
  RectangularSimulationBox& operator=(const RectangularSimulationBox&);

  bool cubicShape;
  
  void AllocateBoxDetails();
  void CopyBoxDetailsToGPU();

 public:
  RectangularSimulationBox();
  RectangularSimulationBox( float Lx, float Ly, float Lz );
  ~RectangularSimulationBox();   

  SimulationBox* MakeCopy();
  void CopyAnotherBox(SimulationBox* boxToCopy);

  __host__ __device__ float4 calculateDistance( float4 position1, float4 position2, float* simBoxPointer ){
    float dr_x = position1.x - position2.x;
    float dr_y = position1.y - position2.y;
    float dr_z = position1.z - position2.z;

    dr_x -= simBoxPointer[0] * rintf( dr_x * simBoxPointer[3] );
    dr_y -= simBoxPointer[1] * rintf( dr_y * simBoxPointer[4] );
    dr_z -= simBoxPointer[2] * rintf( dr_z * simBoxPointer[5] );
    
    float4 value = { dr_x, dr_y, dr_z, dr_x * dr_x + dr_y * dr_y + dr_z * dr_z }; 
    return value;
  }

  __host__ __device__ float4 calculateDistanceWithImages( float4 position1, float4 position2, float4 image1, float4 image2, float* simBoxPointer ){
    float dr_x = (position1.x - position2.x) + simBoxPointer[0] * ( image1.x - image2.x );
    float dr_y = (position1.y - position2.y) + simBoxPointer[1] * ( image1.y - image2.y );
    float dr_z = (position1.z - position2.z) + simBoxPointer[2] * ( image1.z - image2.z );
    
    float4 value = { dr_x, dr_y, dr_z, dr_x * dr_x + dr_y * dr_y + dr_z * dr_z }; 
    return value;
  }
  
  __host__ __device__ float4 applyBoundaryCondition( float4& position, float* simBoxPointer ){
    float4 image = { rintf( position.x * simBoxPointer[3] ), 
		     rintf( position.y * simBoxPointer[4] ), 
		     rintf( position.z * simBoxPointer[5] ), 
		     0.f };
    
    // Periodic boundary condition.
    position.x -= simBoxPointer[0] * image.x;
    position.y -= simBoxPointer[1] * image.y;
    position.z -= simBoxPointer[2] * image.z;
    
    return image; 
  }
  
  __host__ __device__ void loadMemory( float* toMem, float* fromMem, bool isCubicShape ){
    if(isCubicShape){
      toMem[0] = fromMem[0];
      toMem[3] = 1.f / toMem[0]; 
      toMem[1] = toMem[0];
      toMem[2] = toMem[0];
      toMem[4] = toMem[3];
      toMem[5] = toMem[3];
    }
    else{
      for(unsigned int i=0; i < simulationBoxSize; i++)
	toMem[i] = fromMem[i];
    }
  }
  
  void ScaleBox( float factor );

  // Set methods
  void SetBoxDetails( float* details );
  
  // Get methods
  bool GetIsCubicShape(){ return cubicShape; };  
  float GetVolume(){ return h_boxDetails[0] * h_boxDetails[1] * h_boxDetails[2]; }
  float4 GetSimulationBox(){ float4 value = { h_boxDetails[0], h_boxDetails[1], h_boxDetails[2], 0 };  return value; }
};

#endif // SIMULATIONBOX_H
