
/*
  Copyright (C) 2010  Thomas Schrøder
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  LICENSE.txt file for license details.
*/

#include "rumd/BondPotential.h"
#include "rumd/MoleculeData.h"
#include "rumd/Sample.h"

///////////////////////////////////////////////////////
// Harmonic potential
///////////////////////////////////////////////////////

BondHarmonic::BondHarmonic(MoleculeData *globalM, Sample *globalS){
  M = globalM;
  S = globalS;
  SetID_String("bondPot");
}

void BondHarmonic::CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs){
  if ( initialize )
    Harmonic<1, 1, true><<<M->num_blocks, M->threads_per_block>>>
      (M->d_blist, S->particleData.d_r, S->particleData.d_f,  S->particleData.d_w, 
       M->threads_per_block.x, M->max_num_bonds, S->GetNumberOfParticles(),  
       S->GetNumberOfVirtualParticles(), M->SB, M->SB->GetDevicePointer(), 
       M->d_btlist, M->d_bplist, M->d_btlist_int, M->SB->GetIsCubicShape());
  else
    Harmonic<1, 1, false><<<M->num_blocks, M->threads_per_block>>>
      (M->d_blist, S->particleData.d_r, S->particleData.d_f, S->particleData.d_w,  
       M->threads_per_block.x, M->max_num_bonds, S->GetNumberOfParticles(),  
       S->GetNumberOfVirtualParticles(), M->SB, M->SB->GetDevicePointer(), 
       M->d_btlist, M->d_bplist, M->d_btlist_int, M->SB->GetIsCubicShape());
}

double BondHarmonic::GetPotentialEnergy(){
  float4 *f;
  
  Harmonic<1, 2, false><<<M->num_blocks, M->threads_per_block>>>
    (M->d_blist, S->particleData.d_r, S->particleData.d_f, S->particleData.d_w,  
     M->threads_per_block.x, M->max_num_bonds, S->GetNumberOfParticles(),  
     S->GetNumberOfVirtualParticles(), M->SB, M->SB->GetDevicePointer(), 
     M->d_btlist, M->d_bplist, M->d_btlist_int, M->SB->GetIsCubicShape());
  
  // copy to host
  unsigned npart = S->GetNumberOfParticles();
  size_t nbytes = npart*sizeof(float4);
  
  if ( cudaMallocHost((void **)&f, nbytes) == cudaErrorMemoryAllocation )
    throw RUMD_Error("BondHarmonic","GetPotentialEnergy","Memory allocation failure");
  
  cudaMemcpy(f, S->particleData.d_f, nbytes, cudaMemcpyDeviceToHost);
  double e = 0.0;
  for ( unsigned i=0; i<npart; i++ )
    e += f[i].w;
  
  cudaFreeHost(f);
  
  return e;
}

///////////////////////////////////////////////////////
// FENE potential
///////////////////////////////////////////////////////

BondFENE::BondFENE(MoleculeData *globalM, Sample *globalS){
  M = globalM;
  S = globalS;
  SetID_String("bondPot");
}

void BondFENE::CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs){
  if ( initialize )
    FENE<1, 1, true><<<M->num_blocks, M->threads_per_block>>>( M->d_blist, S->particleData.d_r, S->particleData.d_f,  S->particleData.d_w, 
							       M->threads_per_block.x, M->max_num_bonds, S->GetNumberOfParticles(),  
							       S->GetNumberOfVirtualParticles(), M->SB, M->SB->GetDevicePointer(), 
							       M->d_btlist, M->d_btlist_int, M->SB->GetIsCubicShape() );
  else
    FENE<1, 1, false><<<M->num_blocks, M->threads_per_block>>>( M->d_blist, S->particleData.d_r, S->particleData.d_f, S->particleData.d_w,  
								M->threads_per_block.x, M->max_num_bonds, S->GetNumberOfParticles(),  
								S->GetNumberOfVirtualParticles(), M->SB, M->SB->GetDevicePointer(), 
								M->d_btlist, M->d_btlist_int, M->SB->GetIsCubicShape() );
}

double BondFENE::GetPotentialEnergy(){
  float4 *f;
  
  FENE<1, 2, false><<<M->num_blocks, M->threads_per_block>>>( M->d_blist, S->particleData.d_r, S->particleData.d_f, S->particleData.d_w,  
							      M->threads_per_block.x, M->max_num_bonds, S->GetNumberOfParticles(),  
							      S->GetNumberOfVirtualParticles(), M->SB, M->SB->GetDevicePointer(), 
							      M->d_btlist, M->d_btlist_int, M->SB->GetIsCubicShape() );
  
  // copy to host
  unsigned npart = S->GetNumberOfParticles();
  size_t nbytes = npart*sizeof(float4);
  
  if ( cudaMallocHost((void **)&f, nbytes) == cudaErrorMemoryAllocation )
    throw RUMD_Error("BondHarmonic","GetPotentialEnergy","Memory allocation failure");
  
  cudaMemcpy(f, S->particleData.d_f, nbytes, cudaMemcpyDeviceToHost);
  
  double e = 0.0;
  for ( unsigned i=0; i<npart; i++ )
    e +=  f[i].w;
  
  cudaFreeHost(f);
  
  return e;
}

///////////////////////////////////////////////////////
// Kernel implementations of bond potentials
///////////////////////////////////////////////////////

template <int stress, int energy, bool initialize, class Simbox> __global__ void FENE( uint1 *blist, float4 *r, float4 *f, float4 *w,  
										       unsigned tpb, unsigned max_num_bonds, unsigned npart, 
										       unsigned virt_part, Simbox *simbox, float *simboxpointer, 
										       int1 *btlist, int1 *btlist_int, bool shape ){
  
  if( MyGP < npart ){
    // Local variables
    float4 my_f   = { 0.f, 0.f, 0.f, 0.f };
    float my_sts[6] = { 0.f, 0.f, 0.f, 0.f, 0.f, 0.f };
    
    // global atom index
    unsigned ai = blockIdx.x*tpb + threadIdx.x;
    
    // Bonding information offset in blist for ai 
    unsigned offset = virt_part + ai*max_num_bonds;

    // Load the simulation box in local memory to avoid bank conflicts.
    float array[simulationBoxSize];
    simbox->loadMemory(array, simboxpointer, shape);
    
    for ( unsigned n=0; n<blist[ai].x; n++ ){
      int type  = btlist[max_num_bonds*ai+n].x;
      
      if ( btlist_int[type].x == 1 ){
	unsigned bi = blist[offset+n].x;
	
	float4 dr = simbox->calculateDistance(r[ai], r[bi], array);
	
	float rr = dr.w/(1.5f*1.5f);
	float ft = - 30.0f/(1.0f - rr);
	
	my_f.x += ft*dr.x; 
	my_f.y += ft*dr.y; 
	my_f.z += ft*dr.z; 
	
	my_f.w -= logf(1.0f-rr);
	
	// Ready for the off-diagonal components
	my_sts[0] += dr.x*ft*dr.x;
	my_sts[1] += dr.x*ft*dr.y;
	my_sts[2] += dr.x*ft*dr.z;
	my_sts[3] += dr.y*ft*dr.y;
	my_sts[4] += dr.z*ft*dr.y;
	my_sts[5] += dr.z*ft*dr.z;
      }
    }
    
    // 0.5 for the double counting 0.5 for the potential
    // 1.5 for R0 and 30 for k0
    my_f.w *= 0.25f*1.5f*1.5f*30.0f; 
    
    if ( energy == 2 )
      f[ai].w = my_f.w;
    else if ( initialize ){
      f[ai] = my_f;
      if ( stress == 1 ) 
	w[ai].w = my_sts[0] +  my_sts[3] + my_sts[5];
      if ( energy == 1 )
	f[ai].w = my_f.w;
    }
    else {
      f[ai].x += my_f.x;
      f[ai].y += my_f.y;
      f[ai].z += my_f.z;
      f[ai].w += my_f.w;
      if ( stress == 1 ) 
	w[ai].w += my_sts[0] + my_sts[3] + my_sts[5];
      if ( energy == 1 )
	f[ai].w += my_f.w;
    }
  }
}

template <int stress, int energy, bool initialize, class Simbox> __global__ void Harmonic( uint1 *blist, float4 *r, float4 *f, float4 *w,  
											   unsigned tpb, unsigned max_num_bonds, unsigned npart, 
											   unsigned virt_part, Simbox *simbox, float *simboxpointer, 
											   int1 *btlist, float2 *bplist, int1 *btlist_int, bool shape ){
  if( MyGP < npart ){
    // Local variables
    float4 my_f   = { 0.f, 0.f, 0.f, 0.f };
    float my_sts[6] = { 0.f, 0.f, 0.f, 0.f, 0.f, 0.f };

    // global atom index
    unsigned ai = blockIdx.x*tpb + threadIdx.x;
    
    // Bonding information offset in blist for ai 
    unsigned offset = virt_part + ai*max_num_bonds;
    
    // Load the simulation box in local memory to avoid bank conflicts.
    float array[simulationBoxSize];
    simbox->loadMemory(array, simboxpointer, shape);
    
    for ( unsigned n=0; n<blist[ai].x; n++ ){
      int type  = btlist[max_num_bonds*ai+n].x;
      
      if ( btlist_int[type].x == 0 ){
	unsigned bi = blist[offset+n].x;
	float2 p = bplist[type];
	
	float4 dr = simbox->calculateDistance(r[ai], r[bi], array);
	
	float rij = sqrt(dr.w);
	float ft  = -p.y*(1.0f - p.x/rij);
	
	my_f.x += ft*dr.x; 
	my_f.y += ft*dr.y; 
	my_f.z += ft*dr.z; 
	
	my_f.w += p.y*(rij - p.x)*(rij - p.x);
	
	// Ready for the off-diagonal components
	my_sts[0] += dr.x*ft*dr.x;
	my_sts[1] += dr.x*ft*dr.y;
	my_sts[2] += dr.x*ft*dr.z;
	my_sts[3] += dr.y*ft*dr.y;
	my_sts[4] += dr.z*ft*dr.y;
	my_sts[5] += dr.z*ft*dr.z;
      }
    }
    
    // 0.5 for the double counting 0.5 for the potential expression
    my_f.w *= 0.25f; 
    
    if ( energy == 2 )
      f[ai].w = my_f.w;
    else if ( initialize ){
      f[ai] = my_f;
      if ( stress == 1 ) 
	w[ai].w = my_sts[0] + my_sts[3] + my_sts[5];
      if ( energy == 1 )
	f[ai].w = my_f.w;
    }
    else {
      f[ai].x += my_f.x;
      f[ai].y += my_f.y;
      f[ai].z += my_f.z;
      f[ai].w += my_f.w;
      if ( stress == 1 ) 
	w[ai].w += my_sts[0] + my_sts[3] + my_sts[5];
      if ( energy == 1 )
	f[ai].w += my_f.w;
    }
  }  
}
