#include "rumd/EnergiesOutputManager.h"
#include "rumd/Sample.h"
#include "rumd/Potential.h"

#include "rumd/IntegratorNVU.h"
#include "rumd/IntegratorNVT.h"
#include <algorithm>
#include <iostream>

void EnergiesOutputManager::Write() {


  if(logLin.index == 0) // (start a new file)
    {
      if(gz_energyFile) gzclose(gz_energyFile);
      std::ostringstream energyFilename;
      energyFilename << outputDirectory << "/" << baseFilename << std::setfill('0') << std::setw(numFilenameDigits) << std::setprecision(numFilenameDigits) << logLin.block << ".dat.gz";
      gz_energyFile = gzopen(energyFilename.str().c_str(),"w");
      
      // comment line at top of new file
      std::map<std::string, bool>::const_iterator it;
      float dt = 0.;
      if(sample->GetIntegrator())
	dt = sample->GetIntegrator()->GetTimeStep();
      
      
      gzprintf(gz_energyFile, "# ioformat=%d N=%d",energies_ioformat,sample->GetNumberOfParticles());

      if(metaData.has_logLin)
	gzprintf(gz_energyFile," timeStepIndex=%u logLin=%u,%u,%u,%u,%u", logLin.nextTimeStep, logLin.block, logLin.base, logLin.index, logLin.maxIndex, logLin.maxInterval);
      else
	gzprintf(gz_energyFile," Dt=%f", dt*metaData.writeInterval);

      gzprintf(gz_energyFile," columns=");
      bool haveFirst = false;
      // list structure to iterate through and then print only the items which
      // are "true", for the comment line (changes the order)
      
      for(it = metaData.Start(); it != metaData.End(); it++ )
	if(it->second) {
	  if(haveFirst)
	    gzprintf(gz_energyFile, ",");
	  
	  gzprintf(gz_energyFile, "%s",metaData.GetFileString(it->first).c_str());
	  haveFirst = true;
	} // if(it->second)
	gzprintf(gz_energyFile, "\n");
	
	} // end if(logLin.index == 0)  
  WriteOneLine();
  
}


void EnergiesOutputManager::AddExternalCalculator(ExternalCalculator* calc) {
  externalCalculators.push_back(calc);
  calc->GetDataInfo(metaData.options, metaData.fileStr);
}


void EnergiesOutputManager::RemoveExternalCalculator(ExternalCalculator* calc) {
 calc->RemoveDataInfo(metaData.options, metaData.fileStr); 

 std::vector<ExternalCalculator*>::iterator found_calc = std::find(externalCalculators.begin(), externalCalculators.end(), calc);
 if(found_calc == externalCalculators.end())
   throw RUMD_Error("EnergiesOutputManager","AddExternalCalculator","Calculator not found");
 externalCalculators.erase(found_calc);
 
}

void EnergiesOutputManager::WriteOneLine(){

  std::map<std::string, float> Values;
  CalcEnergies(Values);
  
  const IntegratorNVT* testNVT = dynamic_cast<const IntegratorNVT*>(sample->GetIntegrator());
  const IntegratorNVU* testNVU = dynamic_cast<const IntegratorNVU*>(sample->GetIntegrator());


  float4 totalMomentum = { 0, 0, 0, 0 };
  if(testNVT) {
    totalMomentum = testNVT->GetTotalMomentum();
    Values["v_com_x"] = totalMomentum.x / totalMomentum.w;
    Values["v_com_y"] = totalMomentum.y / totalMomentum.w;
    Values["v_com_z"] = totalMomentum.z / totalMomentum.w;
    Values["thermostat_Ps"] = testNVT->GetThermostatState();
  }
  
  if(testNVU){
    Values["simulationDisplacementLength"] = testNVU->GetSimulationDisplacementLength();
    Values["instantaneousTimeStepSq"] = testNVU->GetSimulationTimeStepSq();
    Values["euclideanLength"] = testNVU->GetEuclideanLength();
  }
    
  // iterate over externalCalculators to get their data-values
  for(unsigned int edx = 0; edx < externalCalculators.size(); edx++) 
    externalCalculators[edx]->GetDataValues(Values);

  int precision = metaData.precision;

  // iterate over all items and write them to the current line
  std::map<std::string, bool>::const_iterator it;
  for(it = metaData.Start(); it != metaData.End(); it++ )
    if(it->second) {
      gzprintf(gz_energyFile, "%.*g ", precision, Values[it->first]);
    }
  // final newline
  gzprintf(gz_energyFile, "\n");
}


void EnergiesOutputManager::CalcEnergies(std::map<std::string, float>& dataValues)
{
  unsigned int degreesOfFreedom = sample->GetNumberOfDOFs();
  const Integrator* itg = sample->GetIntegrator();
  const ParticleData& particleData = sample->GetParticleData();
  unsigned int N = sample->GetNumberOfParticles();  
  float V = sample->GetSimulationBox()->GetVolume();

   
  float4 sumC = { 0, 0, 0, 0 };
  float stress[6] = { 0, 0, 0, 0, 0, 0 };

  double stress_tmp[6] = {0., 0., 0., 0., 0., 0.};
  double2 config_temp = { 0, 0 };
  double constraintW = 0;

  float4* h_f = particleData.h_f;
  float4* h_w = particleData.h_w;
  float4* h_sts = particleData.h_sts;

  bool need_stresses = metaData.Get("stress_xx") || metaData.Get("stress_yy") || metaData.Get("stress_zz") || metaData.Get("stress_xy") || metaData.Get("stress_yz") || metaData.Get("stress_xz");
  if(need_stresses)
    sample->CalcF(true);

  // maybe check using counters, and only copy specifically what's needed?
  particleData.CopyConfFromDevice();

  for ( unsigned int i=0; i < N; i++ ){ 
    stress_tmp[0] += h_sts[i].x;
    stress_tmp[1] += h_sts[i].y;
    stress_tmp[2] += h_sts[i].z;
    stress_tmp[3] += h_sts[i].w;
    stress_tmp[4] += h_w[i].y;
    stress_tmp[5] += h_w[i].z;
    
    // Nom and denom for configurational temperature.
    config_temp.x += h_f[i].x * h_f[i].x + h_f[i].y * h_f[i].y + h_f[i].z * h_f[
										i].z;
    config_temp.y += particleData.h_misc[i].x;
    constraintW += particleData.h_misc[i].y;
  }  
  
  // Temperature
  double ke = 0.0;
  if(itg) ke = itg->GetKineticEnergy(); // includes copy
  float T = ( 2.0 * ke ) / (double) degreesOfFreedom;
  ke = float(ke);
  sumC.x = config_temp.x;
  sumC.y = -config_temp.y;


  float pe = sample->GetPotentialEnergy(); // includes copy  
  float vir = sample->GetVirial();
  // Constraint virial.
  sumC.z = constraintW / 3.0;


  // individual potential energy contributions
  std::vector<Potential*> *potentials = sample->GetPotentials();
  unsigned int numPotentials = potentials->size();
  if(numPotentials)
    for(unsigned int pdx = 0; pdx < numPotentials; pdx++) {
      std::string potID = (*potentials)[pdx]->GetID_String();
      if(metaData.Get(potID))
	dataValues[potID] = (*potentials)[pdx]->GetPotentialEnergy()/N;
  }


  // Total stress
  for(unsigned int j = 0; j < 6; j++)
    stress[j] = 0.5*stress_tmp[j];


  float E = pe + ke;
  float W = vir + sumC.z;
  float P = ( N * T + W ) / V;
  float H = E + P * V; 
  

  dataValues["kineticEnergy"] = ke / N;
  dataValues["potentialEnergy"] = pe / N;
  dataValues["totalEnergy"] = E / N;
  dataValues["virial"] = W / N;
  dataValues["potentialVirial"] = vir / N;
  dataValues["constraintVirial"] = sumC.z / N;
  dataValues["temperature"] = T;
  dataValues["configTemperature"] = sumC.x / sumC.y; 
  dataValues["pressure"] = P; 
  dataValues["volume"] = V;
  dataValues["stress_xx"] = stress[0]/V;
  dataValues["stress_yy"] = stress[1]/V;
  dataValues["stress_zz"] = stress[2]/V;
  dataValues["stress_yz"] = stress[3]/V;
  dataValues["stress_xz"] = stress[4]/V;
  dataValues["stress_xy"] = stress[5]/V;
  dataValues["enthalpy"] = H / N;
  dataValues["laplacian"] = sumC.y; 
}
